#include "alarm-manager.h"
#include "ui_alarm-manager.h"

// cumbia
#include <cumbiapool.h>
#include <cuserviceprovider.h>
#include <cumacros.h>
#include <quapps.h>
// cumbia

// tree model plugin
#include <qutreemodelplugininterface.h>

#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QLabel>
#include <quwatcher.h>
#include <qulineedit.h>
#include <qulabel.h>
#include <qubutton.h>
#include <QGridLayout>
#include <QAbstractItemModel>
#include <QTreeView>
#include <QSplitter>
#include <QLineEdit>
#include <cupluginloader.h>
#include <QGroupBox>
#include <qscrollarea.h>
#include <QtDebug>
#include <quinputoutput.h>
#include <qustringlist.h>
#include <quwriter.h>
#include <qucheckbox.h>
#include <QMessageBox>
#include <qustring.h>
#include <qtimer.h>

Alarmmanager::Alarmmanager(CumbiaPool *cumbia_pool, QWidget *parent) :
    QWidget(parent)
{
    // cumbia
    CuModuleLoader mloader(cumbia_pool, &m_ctrl_factory_pool, &m_log_impl);
    cu_pool = cumbia_pool;

    CuPluginLoader loader;
    QObject **qo = nullptr;
    tree_i = loader.get<QuTreeModelPluginInterface>("libqutreemodel-plugin.so", qo);
    if(!tree_i)
        perr("%s: plugin \"%s\" not found", __FUNCTION__, "libqutreemodel-plugin.so");
    else {
        QuTreeModelI *mi = tree_i->instantiate({"Alarm"}, this);
        m_model = mi->qmodel();
        m_model_i = mi;

        //    ui = new Ui::Alarmmanager;
        //    ui->setupUi(this, cu_pool, m_ctrl_factory_pool);

        // mloader.modules() to get the list of loaded modules
        // cumbia
        m_load_alarm_list(); // if ok, we will find at least one Qu1TWatcher

        // use one of the devices to get the list of attributes
        QList<Qu1TWatcher *> was = findChildren<Qu1TWatcher *>();
        if(was.size() > 0) {
            // just take attributes from one dev
            Qu1TWatcher *aw = new Qu1TWatcher(this, cumbia_pool, m_ctrl_factory_pool);
            connect(aw, SIGNAL(newData(CuData)), this, SLOT(m_load_atts(CuData)));
            aw->setSource(was.first()->source().section('/', 0, 2) + "/");
        }
        else
            perr("no watchers found");
        QGridLayout *lo = new QGridLayout(this);
        lo->setObjectName("lo0");
        // splitter between view and scroll area with contents
        QSplitter *spli = new QSplitter(Qt::Horizontal, this);
        // widget for view + search
        QWidget *vw = new QWidget(this);
        QGridLayout *wlo = new QGridLayout(vw);
        // tree view
        QTreeView *t = new QTreeView(vw);
        wlo->addWidget(t, 0, 0, 10, 4);
        t->setSelectionMode(QAbstractItemView::SingleSelection);
        t->setModel(this->m_model_i->qmodel());
        connect(t->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)),
                this, SLOT(m_selection_changed(QItemSelection,QItemSelection)));

        QLabel *search_l = new QLabel("Search", vw);
        QLineEdit *search_le = new QLineEdit(vw);
        search_le->setObjectName("search_le");
        search_le->setPlaceholderText("Type text to search...");
        connect(search_le, SIGNAL(textChanged(QString)), this, SLOT(m_search_text_changed(QString)));
        wlo->addWidget(search_l, 10, 0, 1, 1);
        wlo->addWidget(search_le, 10, 1, 1, 3);

        spli->addWidget(vw);

        QScrollArea *scroll = new QScrollArea(this);
        QGroupBox *gb = new QGroupBox("Edit alarms", this);
        gb->setObjectName("gbEdit");
        // layout for group box
        QGridLayout *gblo = new QGridLayout(gb);
        gblo->setObjectName("gblo");
        scroll->setWidgetResizable(true);
        scroll->setWidget(gb);
        spli->addWidget(scroll);
        spli->setStretchFactor(1, 2);

        lo->addWidget(spli, 0, 0, 10, 10);

        resize(1024, 768);
    }
}

Alarmmanager::~Alarmmanager() {
    //    delete ui;
}

void Alarmmanager::m_search_text_changed(const QString& ) {
    QTimer *t = findChild<QTimer *>("search_text_changed_tmr");
    if(!t) {
        t = new QTimer(this);
        t->setSingleShot(true);
        t->setObjectName("search_text_changed_tmr");
        connect(t, SIGNAL(timeout()), this, SLOT(m_filter()));
    }
    t->start(1000);
}

void Alarmmanager::m_filter() {
    const int min_pattern_len = 2;
    const QString& s = findChild<QLineEdit *>("search_le")->text();
//    if(s.size() > 1) {
        QList<QModelIndex > il = m_model_i->findItems(s, min_pattern_len, findChild<QTreeView *>());
        QModelIndex last = il.size() > 0 ? il.last() : QModelIndex();
        if(last.isValid())
            findChild<QTreeView *>()->scrollTo(last);
//    }
}

void Alarmmanager::m_load(const CuData& da, bool expand) {
    if(da.B(CuDType::Err)) {  // da.B("err")
        QMessageBox::information(this, "Setup error", da.s(CuDType::Message).c_str());  // da.s("msg")
    }
    else {
        QModelIndex newitemidx;
        QStringList before = m_model_i->items();
        pretty_pri("before list was");
        for(const QString& a : before)
            printf("'%s', ", qstoc(a));
        printf("\n\n");
        QuStringList args = da[CuDType::Value].toStringVector();  // da["value"]
        //        qDebug() << __PRETTY_FUNCTION__ << args << sender();
        foreach(QString s, args) {
            s.remove(QRegularExpression("tango://[A-Za-z0-9\\.\\-:]+/"));
            QModelIndex mi = m_model_i->addItem(s);
            this->m_model->setData(mi, sender()->property("device"), Qt::UserRole);
            if(before.size() > 0 && !before.contains(s))
                newitemidx = mi;
        }
        if(expand)
            m_expand_l3();
        pretty_pri("expanding item '%s'", qstoc(newitemidx.data().toString()));
        m_model_i->expand(newitemidx, findChild<QTreeView *>());
        m_model_i->highlight(newitemidx);
    }
}

void Alarmmanager::m_load_atts(const CuData &da) {
    if(da.B(CuDType::Err)) {  // da.B("err")
        perr("%s: error loading attributes: %s", __PRETTY_FUNCTION__, da.s(CuDType::Err).c_str());  // da.s("err")
    } else {
        // right pane with attributes
        QGroupBox *gb = findChild<QGroupBox *>("gbEdit");
        QGridLayout *lo = gb->findChild<QGridLayout *>("gblo");
        QuStringList atts = da[CuDType::Value].toStringVector();  // da["value"]
        QStringList excl { "State", "Status", "alarmList", "sync_write", "default_archiver" };
        for(int i = 0; i < atts.size(); i++) {
            const QString& a = atts[i];
            if(a == "handlerStatus") {

            }
            else if(a == "alarmFrequency") {

            }
            else if(a == "enabled") {
                QuCheckBox *qcb = new QuCheckBox(this, cu_pool, m_ctrl_factory_pool);
                qcb->setProperty("attribute", a);
                lo->addWidget(qcb, i, 1, 1, 2);
            }
            else if(!excl.contains(a)) {
                QLabel *l = new QLabel(a, gb);
                lo->addWidget(l, i, 0, 1, 1);
                QuInputOutput *qio = new QuInputOutput(this, cu_pool, m_ctrl_factory_pool);
                qio->setProperty("attribute", a);
                lo->addWidget(qio, i, 1, 1, 6);
            }
        }
        // Load, Modify, AddArchiver and Remove commands
        const int r = lo->rowCount();
        int col = 0;
        const QStringList cmds {"Load", "Modify", "AddArchiver" , "Remove" };
        for( int i = 0; i < cmds.size(); i++) {
            QuButton *butt = new QuButton(gb, cu_pool, m_ctrl_factory_pool, cmds[i]);
            butt->setObjectName("b" + cmds[i]);
            butt->setProperty("cmd", cmds[i]);
            lo->addWidget(butt, r, col++, 1, 1);
            connect(butt, SIGNAL(newData(CuData)), this, SLOT(m_on_change(CuData)));
        }
        // prevent involuntary "Remove"
        QCheckBox *cb = new QCheckBox("Enable remove", this);
        connect(cb, SIGNAL(toggled(bool)), gb->findChild<QuButton *>("bRemove"), SLOT(setEnabled(bool)));
        connect(gb->findChild<QuButton *>("bRemove"),
                SIGNAL(clicked()), cb, SLOT(toggle()));
        cb->setChecked(false);
        gb->findChild<QuButton *>("bRemove")->setEnabled(false);
        lo->addWidget(cb, r, col++, 1, 1);
    }
}

void Alarmmanager::m_selection_changed(const QItemSelection &sel, const QItemSelection &) {
    pretty_pri("selection indexes size %lld", sel.indexes().size());
    for(int i = 0; i < sel.indexes().size(); i++)
        printf("- %s\n", qstoc(sel.indexes().at(i).data().toString()));
    if(sel.indexes().size() > 0) {
        QGroupBox *g = findChild<QGroupBox *>("gbEdit");
        const QModelIndex mi = sel.indexes().first();
        const QString& n = m_model_i->itemRepr(mi);
        const QString& dev = mi.data(Qt::UserRole).toString();
        m_model_i->clearHighlight(mi);
        if(dev.count('/') >= 2) {
            printf("tag: %s / device %s\n", qstoc(n), qstoc(mi.data(Qt::UserRole).toString()));
            // write on tag
            QuWriter *w  = findChild<QuWriter *>("tag_writer");
            if(!w) {
                w = new QuWriter(this, cu_pool, m_ctrl_factory_pool);
                connect(w, SIGNAL(newData(CuData)), this, SLOT(m_on_tag_writer_data(CuData)));
                w->setObjectName("tag_writer");
            }
            w->setProperty("device", dev);
            w->setProperty("tag", n);
            w->setTarget(dev + "/tag");
            printf("Alarmmanager::m_selection_changed: \e[1;31mexecuting %s(%s)\e[0m writer %p\n",
                   qstoc(w->target()), qstoc(n), w);
            w->execute(n);
        }
    }
}

void Alarmmanager::m_on_tag_writer_data(const CuData &da) {
    if(da.B(CuDType::Err)) {  // da.B("err")
        QMessageBox::information(this, "Error updating tag", da.s(CuDType::Message).c_str());  // da.s("msg")
    } else if(da.B("is_result")) {
        printf("Alarmmanager::m_on_tag_writer_data: data \e[1;33m%s\e[0m\n", datos(da));
        m_tag_written(da.s(CuDType::Device).c_str());  // da.s("device")
    }
}

void Alarmmanager::m_tag_written(const QString& dev) {
    QGroupBox *g = findChild<QGroupBox *>("gbEdit");
    printf("\e[1;32m refreshing sources for dev %s \e[0m\n", qstoc(dev));
    foreach(QuInputOutput *io, g->findChildren<QuInputOutput *>()) {
        io->unsetSource();
        io->setSource(dev + "/" + io->property("attribute").toString());
    }
    foreach(QuCheckBox *cb,  g->findChildren<QuCheckBox *>()) {
        cb->setSource(dev + "/" + cb->property("attribute").toString());
    }
    foreach(QuButton *b, g->findChildren<QuButton *>(QRegularExpression("b.*"))) {
        // on command buttons update the target only if device changes. Use a regexp to capture
        // current device on buttons target.
        // When the application is used with a single device, target will be set once, only when t
        // is empty
        QString t = b->target();
        static QRegularExpression re("(?:[a-zA-Z0-9\\.\\-_]+:\\d+/){0,1}(.*)\\-\\>[a-zA-Z0-9\\.\\-_]+");
        QRegularExpressionMatch ma = re.match(t);
        qDebug() << __PRETTY_FUNCTION__ << "button" << b->objectName() <<  "captured texts" << ma.capturedTexts()
                 << "target " << t << "regex " << re.pattern();
        if(t.isEmpty() || (ma.capturedTexts().size() == 2 && ma.capturedTexts().at(1) != dev)) {
            t = dev + "->" + b->property("cmd").toString();
            printf("\e[1;31msetting target on button %p '%s': %s (old target '%s')\e[0m\n",
                   b, qstoc(b->objectName()), qstoc(t), qstoc(b->target()));
            b->setTarget(t);
        }
    }
}


void Alarmmanager::m_expand_l3() {
    QTreeView *v = findChild<QTreeView* >();
    QAbstractItemModel *m = m_model_i->qmodel();
    for(int i1 = 0; i1 < m->rowCount(); i1++) {
        const QModelIndex& mi = m->index(i1, 0);
        v->expand(mi);
        for(int i2 = 0; i2 < m->rowCount(mi); i2++) {
            const QModelIndex& mi2 = m->index(i2, 0, mi);
            v->expand(mi2);
        }
    }
}

void Alarmmanager::m_on_change(const CuData& da) {
    qDebug() << __PRETTY_FUNCTION__ << sender();
    printf("\e[1;35mAlarmmanager::m_on_change %s\e[0m\n", datos(da));
    bool e = da.B(CuDType::Err);  // da.B("err")
    bool is_res = da.b("is_result");
    const QString &on = sender()->objectName();
    const QString &src = findChild<QuWriter *>("tag_writer")->property("tag").toString();
    if(!e && is_res && ( on == "bModify" || on == "bAddArchiver")) {
        printf("Alarmmanager::m_on_change modified '%s'\n", qstoc(src));
        m_model_i->select(src, findChild<QTreeView *>());
    }
    else if(!e && is_res && findChild<QuWriter *>("tag_writer") != nullptr) {
        pretty_pri("data %s \e[1;32m tag %s\e[0m", datos(da), qstoc(src));
        if(QuString(da.s(CuDType::Src)).endsWith("Remove") && !src.isEmpty()) {  // da.s("src")
            QItemSelectionModel *sm = findChild<QTreeView *>()->selectionModel();
            sm->clearSelection();
            printf("removing item %s\n", qstoc(src));
            QModelIndex upi = m_model_i->removeItem(src);
//            sm->select(upi, QItemSelectionModel::Select);
        }
        else {
            m_load_alarm_list();
        }
    }
    if(e)
        QMessageBox::critical(this, "Error executing " + QuString(da.s(CuDType::Src)), QuString(da.s(CuDType::Message)));  // da.s("src"), da.s("msg")
}

void Alarmmanager::m_load_alarm_list() {
    pretty_pri("Alarmmanager::m_load_alarm_list(): enter");
    for(int i = 1; i < qApp->arguments().size(); i++) {
        const QString s = qApp->arguments().at(i);
        if(s.count('/') >= 2) { // probably a device
            Qu1TWatcher *w = new Qu1TWatcher(this, cu_pool, m_ctrl_factory_pool);
            connect(w, SIGNAL(newData(CuData)), this, SLOT(m_load(CuData)));
            w->setSource(s + "/alarmList");
            w->setProperty("device", s);
            pretty_pri("Alarmmanager::m_load_alarm_list() source %s", qstoc( w->source()));
        }
    }
}
